/*
 * PstricksOutdata.cpp - Version 0.4 - May 11, 2019
 *
 * Author: Sam Buss
 *
 * Software accompanying POSSIBLE SECOND EDITION TO the book
 *		3D Computer Graphics: A Mathematical Introduction with OpenGL,
 *		by S. Buss, Cambridge University Press, 2003.
 *
 * Software is "as-is" and carries no warranty.  It may be used without
 *   restriction, but if you modify it, please change the filenames to
 *   prevent confusion between different versions.
 * Bug reports: Sam Buss, sbuss@ucsd.edu.
 * Web page: http://math.ucsd.edu/~sbuss/MathCG2
 */

// Purpose: Reads data from the feedback transform buffer.
//  Each entry in the feedback buffer is 7 values:
//
//   int backFace; // 0 if front facing, 1 if back facing
//   int vertexID1;  // Vertex ID's are 0-based 
//   int vertexID2;
//   float x1, y1;      // Position of the first vertex
//   float x2, y2;      // Position of the second vertex

#define _CRT_SECURE_NO_WARNINGS     // Allow use of fopen.

#include "PstricksOutdata.h"
#include "MathMisc.h"

#include <stdio.h>
#include <assert.h>
#include <limits>

// The FeedbackData_PS holds the data that will be output as part
// latex commands (for pstricks).
// The "priorityCode" serves dual purpose. First, it identifies
//   the type of edge, and in particular how it is to be drawn.
//   Second, when the same edge is present as different types (for
//   instance, part of both a front-facing triangle and a back-facing
//   triangle), then the lower priority version of the edge is used.
// For now, the priorityCodeis even is on front facing triangle, odd otherwise
// We always have vertexID1 < vertexID2.
struct FeedbackData_PS {
    int priorityCode; // The edge type
    int vertexID1;    // ID of the first vertex of the edge
    int vertexID2;    // ID of the second vertex of the edge
    float x1, y1;     // Position of the first vertex;
    float x2, y2;     // Position of the second vertex;
};

// This class is used for sorting edges by priorityCode
class BestEdge {
public:
    int priorityCode;
    int edgeNumber;

    BestEdge() : priorityCode(INT_MAX) {}

    bool operator<(BestEdge b) { 
        return (priorityCode < b.priorityCode || (priorityCode == b.priorityCode && edgeNumber < b.edgeNumber));
    }
};

template<class T> void SwapValues(T& a, T& b) {
    T temp = a;
    a = b;
    b = temp;
}

// Returns value L such that (i,j) is the L-th
// ordered pair with 0 <= i < j.  (Starting with L=0.)
int DiagIndex(int i, int j) {
    assert(0 <= i && i < j);
    int index = ((i - 1)*i / 2) + j;    // Sum of k's < i, then plus j.
    return index;
}

void OutputPstricksData(void* feedbackData, int numEdges, int XformMode)
{
    FeedbackData_PS* theData = (FeedbackData_PS*)feedbackData;

    // Reorder vertices so lowest vertexID is first.
    // Find max vertex number
    float minX = std::numeric_limits<float>::max();
    float maxX = -minX;
    float maxY = maxX;
    float minY = minX;
    int minID = std::numeric_limits<int>::max();
    int maxID = std::numeric_limits<int>::min();
    int minPriority = minID;
    int maxPriority = maxID;
    for (int i = 0; i < numEdges; i++) {
        FeedbackData_PS& t = theData[i];
        assert(t.vertexID1 != t.vertexID2);
        if (t.vertexID1 > t.vertexID2) {
            SwapValues(t.vertexID1, t.vertexID2);
            SwapValues(t.x1, t.x2);
            SwapValues(t.y1, t.y2);
        }
        ClampMax(&minID, t.vertexID1);   // minID = min(minID, t.vertexID1)
        ClampMin(&maxID, t.vertexID2);   // maxID = max(maxID, t.vertexID2)
        ClampMax(&minPriority, t.priorityCode);
        ClampMin(&maxPriority, t.priorityCode);
        ClampMin(&maxX, t.x1);
        ClampMin(&maxX, t.x2);
        ClampMax(&minX, t.x1);
        ClampMax(&minX, t.x2);
        ClampMin(&maxY, t.y1);
        ClampMin(&maxY, t.y2);
        ClampMax(&minY, t.y1);
        ClampMax(&minY, t.y2);
    }
    assert(minID >= 0);

    // Allocate an array to store info on edges.
    //    Indexing is done diagonally (saves memory)
    int N = DiagIndex(maxID-1, maxID);
    BestEdge* edgeInfo = new BestEdge[N];
    assert(edgeInfo != 0);
 
    // Fill in the array entries with lowest priorityCode
    for (int i = 0; i < numEdges; i++) {
        FeedbackData_PS& t = theData[i];
        int idx = DiagIndex(t.vertexID1, t.vertexID2);
        if (edgeInfo[idx].priorityCode > t.priorityCode) {
            edgeInfo[idx].priorityCode = t.priorityCode;
            edgeInfo[idx].edgeNumber = i;
        }
    }

    // Output best edges
    char filename[] = "dataFileN.tex\0";
    filename[8] = '0' + XformMode;
    FILE* outFile = fopen(filename, "w");
    if (!outFile) {
        fprintf(stderr, "Unable to open file %s. \n", filename);
        return;
    }

    assert(XformMode <= 3 && XformMode > 0);
    assert(maxPriority < 4 && minPriority >= 0);
    const char* xformStrings[3] = {
        "SphereSlice", "SphereCap", "TorusSlice"
    };
    const char* priorityStrings[4] = {
        "MainFront", "MainBack", "RestFront", "RestBack"
    };

    for (int pc = minPriority; pc <= maxPriority; pc++) {
        for (int i = 0; i < numEdges; i++) {
            FeedbackData_PS& t = theData[i];
            int idx = DiagIndex(t.vertexID1, t.vertexID2);
            if (edgeInfo[idx].priorityCode == pc) {
                fprintf(outFile, "\\psline[style=%s%s](%0.4f,%0.4f)(%0.4f,%0.4f)\n",
                    xformStrings[XformMode-1], priorityStrings[pc], t.x1, t.y1, t.x2, t.y2);
                edgeInfo[idx].priorityCode = INT_MAX;                // Don't use this edge again
            }
        }
    }
    fclose(outFile);
}

