
// A vertex shader and a geometry shader, not a fragment shader!
//
// These are very basic shaders but support the use 
//     of projection and modelview matrices.
// Their purpose to output data to a Transform Feedback Buffer,
//     which the C++ program can read from.
//

// ***************************
// Vertex Shader, named "vs_XformFeedbackChap1"
//     Sets the position and vertexID of a vertex.
//   The projection and modelview matrices are used to position the vertex.
// ***************************
#beginglsl vertexshader vs_XformFeedbackChap1
#version 330 core
layout (location = 0) in vec3 vPos;	   // Position in attribute location 0
uniform mat4 projectionMatrix;         // The projection matrix    
uniform mat4 modelviewMatrix;          // The model-view matrix    

out int vertexId;				// Vertex's index in the VBO

void main()    
{    
   gl_Position = projectionMatrix * modelviewMatrix * vec4(vPos.x, vPos.y, vPos.z, 1.0);    
   vertexId = gl_VertexID;
}
#endglsl 

// ***************************
// Geometry shader for feedback transform
// ***************************
#beginglsl geometryshader gs_XformFeedbackChap1
// Computes values for Transform Feedback for chapter 1 figures.
#version 330 core  
layout ( triangles ) in;
layout ( points, max_vertices = 3) out;

in int vertexId[];		// Three vertex ID's
uniform int mode;		// An even number 0 or 2

out int drawMode;				// mode (if front-facing) or mode+1 (if back-facing);
out int vertexNumber1;
out int vertexNumber2;
out float xCoord1;
out float yCoord1;
out float xCoord2;
out float yCoord2;

bool isFrontFacing();
vec2 vPos[3];

void main()    
{   
	for (int i=0; i<3; i++ ) {
	   float w = gl_in[i].gl_Position.w;
	   vPos[i].x = gl_in[i].gl_Position.x / w;
	   vPos[i].y = gl_in[i].gl_Position.y / w;
	}
    int frontFacing = isFrontFacing() ? 0 : 1;
    for ( int i=0; i<3; i++ ) {
		drawMode = mode + frontFacing;
		int j = (i+1)%3;
		xCoord1 = vPos[i].x;
		yCoord1 = vPos[i].y;
		xCoord2 = vPos[j].x;
		yCoord2 = vPos[j].y;
		vertexNumber1 = vertexId[i];
		vertexNumber2 = vertexId[j];
		EmitVertex();    // Each call to EmitVertex generates a transform feedback record
    }
	EndPrimitive();
}

// Compute whether front-facing by doing a crossproduct (actually, a determinant).
bool isFrontFacing() {
    float dx1 = vPos[1].x - vPos[0].x;
    float dy1 = vPos[1].y - vPos[0].y;
    float dx2 = vPos[2].x - vPos[0].x;
    float dy2 = vPos[2].y - vPos[0].y;
	float crossP = dx1*dy2 - dy1*dx2;
	return ( crossP>=0.0 );	
}
#endglsl

